@@ -7,7 +7,6 @@ const net = require('net')
 const util = require('./core/util')
 const Request = require('./core/request')
 const DispatcherBase = require('./dispatcher-base')
-const RedirectHandler = require('./handler/redirect')
 const {
   RequestContentLengthMismatchError,
   ResponseContentLengthMismatchError,
@@ -18,7 +17,8 @@ const {
   SocketError,
   InformationalError,
   BodyTimeoutError,
-  HTTPParserError
+  HTTPParserError,
+  ResponseExceededMaxSizeError
 } = require('./core/errors')
 const buildConnector = require('./core/connect')
 const {
@@ -60,7 +60,10 @@ const {
   kCounter,
   kClose,
   kDestroy,
-  kDispatch
+  kDispatch,
+  kInterceptors,
+  kLocalAddress,
+  kMaxResponseSize
 } = require('./core/symbols')
 
 const kClosedResolve = Symbol('kClosedResolve')
@@ -82,6 +85,7 @@ try {
 
 class Client extends DispatcherBase {
   constructor (url, {
+    interceptors,
     maxHeaderSize,
     headersTimeout,
     socketTimeout,
@@ -101,7 +105,9 @@ class Client extends DispatcherBase {
     maxCachedSessions,
     maxRedirections,
     connect,
-    maxRequestsPerClient
+    maxRequestsPerClient,
+    localAddress,
+    maxResponseSize
   } = {}) {
     super()
 
@@ -169,6 +175,14 @@ class Client extends DispatcherBase {
       throw new InvalidArgumentError('maxRequestsPerClient must be a positive number')
     }
 
+    if (localAddress != null && (typeof localAddress !== 'string' || net.isIP(localAddress) === 0)) {
+      throw new InvalidArgumentError('localAddress must be valid string IP address')
+    }
+
+    if (maxResponseSize != null && (!Number.isInteger(maxResponseSize) || maxResponseSize < -1)) {
+      throw new InvalidArgumentError('maxResponseSize must be a positive number')
+    }
+
     if (typeof connect !== 'function') {
       connect = buildConnector({
         ...tls,
@@ -179,6 +193,9 @@ class Client extends DispatcherBase {
       })
     }
 
+    this[kInterceptors] = interceptors && interceptors.Client && Array.isArray(interceptors.Client)
+      ? interceptors.Client
+      : [createRedirectInterceptor({ maxRedirections })]
     this[kUrl] = util.parseOrigin(url)
     this[kConnector] = connect
     this[kSocket] = null
@@ -189,6 +206,7 @@ class Client extends DispatcherBase {
     this[kKeepAliveTimeoutThreshold] = keepAliveTimeoutThreshold == null ? 1e3 : keepAliveTimeoutThreshold
     this[kKeepAliveTimeoutValue] = this[kKeepAliveDefaultTimeout]
     this[kServerName] = null
+    this[kLocalAddress] = localAddress != null ? localAddress : null
     this[kResuming] = 0 // 0, idle, 1, scheduled, 2 resuming
     this[kNeedDrain] = 0 // 0, idle, 1, scheduled, 2 resuming
     this[kHostHeader] = `host: ${this[kUrl].hostname}${this[kUrl].port ? `:${this[kUrl].port}` : ''}\r\n`
@@ -198,6 +216,7 @@ class Client extends DispatcherBase {
     this[kMaxRedirections] = maxRedirections
     this[kMaxRequests] = maxRequestsPerClient
     this[kClosedResolve] = null
+    this[kMaxResponseSize] = maxResponseSize > -1 ? maxResponseSize : -1
 
     // kQueue is built up of 3 sections separated by
     // the kRunningIdx and kPendingIdx indices.
@@ -254,11 +273,6 @@ class Client extends DispatcherBase {
   }
 
   [kDispatch] (opts, handler) {
-    const { maxRedirections = this[kMaxRedirections] } = opts
-    if (maxRedirections) {
-      handler = new RedirectHandler(this, maxRedirections, opts, handler)
-    }
-
     const origin = opts.origin || this[kUrl].origin
 
     const request = new Request(origin, opts, handler)
@@ -319,6 +333,7 @@ class Client extends DispatcherBase {
 }
 
 const constants = require('./llhttp/constants')
+const createRedirectInterceptor = require('./interceptor/redirectInterceptor')
 const EMPTY_BUF = Buffer.alloc(0)
 
 async function lazyllhttp () {
@@ -426,6 +441,7 @@ class Parser {
 
     this.keepAlive = ''
     this.contentLength = ''
+    this.maxResponseSize = client[kMaxResponseSize]
   }
 
   setTimeout (value, type) {
@@ -542,19 +558,6 @@ class Parser {
     }
   }
 
-  finish () {
-    try {
-      try {
-        currentParser = this
-      } finally {
-        currentParser = null
-      }
-    } catch (err) {
-      /* istanbul ignore next: difficult to make a test case for */
-      util.destroy(this.socket, err)
-    }
-  }
-
   destroy () {
     assert(this.ptr != null)
     assert(currentParser == null)
@@ -783,7 +786,7 @@ class Parser {
   }
 
   onBody (buf) {
-    const { client, socket, statusCode } = this
+    const { client, socket, statusCode, maxResponseSize } = this
 
     if (socket.destroyed) {
       return -1
@@ -802,6 +805,11 @@ class Parser {
 
     assert(statusCode >= 200)
 
+    if (maxResponseSize > -1 && this.bytesRead + buf.length > maxResponseSize) {
+      util.destroy(socket, new ResponseExceededMaxSizeError())
+      return -1
+    }
+
     this.bytesRead += buf.length
 
     try {
@@ -917,7 +925,7 @@ function onSocketError (err) {
   // to the user.
   if (err.code === 'ECONNRESET' && parser.statusCode && !parser.shouldKeepAlive) {
     // We treat all incoming data so for as a valid response.
-    parser.finish()
+    parser.onMessageComplete()
     return
   }
 
@@ -951,7 +959,7 @@ function onSocketEnd () {
 
   if (parser.statusCode && !parser.shouldKeepAlive) {
     // We treat all incoming data so far as a valid response.
-    parser.finish()
+    parser.onMessageComplete()
     return
   }
 
@@ -961,6 +969,11 @@ function onSocketEnd () {
 function onSocketClose () {
   const { [kClient]: client } = this
 
+  if (!this[kError] && this[kParser].statusCode && !this[kParser].shouldKeepAlive) {
+    // We treat all incoming data so far as a valid response.
+    this[kParser].onMessageComplete()
+  }
+
   this[kParser].destroy()
   this[kParser] = null
 
@@ -1020,7 +1033,8 @@ async function connect (client) {
         hostname,
         protocol,
         port,
-        servername: client[kServerName]
+        servername: client[kServerName],
+        localAddress: client[kLocalAddress]
       },
       connector: client[kConnector]
     })
@@ -1033,7 +1047,8 @@ async function connect (client) {
         hostname,
         protocol,
         port,
-        servername: client[kServerName]
+        servername: client[kServerName],
+        localAddress: client[kLocalAddress]
       }, (err, socket) => {
         if (err) {
           reject(err)
@@ -1076,7 +1091,8 @@ async function connect (client) {
           hostname,
           protocol,
           port,
-          servername: client[kServerName]
+          servername: client[kServerName],
+          localAddress: client[kLocalAddress]
         },
         connector: client[kConnector],
         socket
@@ -1093,7 +1109,8 @@ async function connect (client) {
           hostname,
           protocol,
           port,
-          servername: client[kServerName]
+          servername: client[kServerName],
+          localAddress: client[kLocalAddress]
         },
         connector: client[kConnector],
         error: err
@@ -1278,7 +1295,7 @@ function _resume (client, sync) {
 }
 
 function write (client, request) {
-  const { body, method, path, host, upgrade, headers, blocking } = request
+  const { body, method, path, host, upgrade, headers, blocking, reset } = request
 
   // https://tools.ietf.org/html/rfc7231#section-4.3.1
   // https://tools.ietf.org/html/rfc7231#section-4.3.2
@@ -1346,7 +1363,6 @@ function write (client, request) {
 
   if (method === 'HEAD') {
     // https://github.com/mcollina/undici/issues/258
-
     // Close after a HEAD request to interop with misbehaving servers
     // that may send a body in the response.
 
@@ -1360,6 +1376,10 @@ function write (client, request) {
     socket[kReset] = true
   }
 
+  if (reset) {
+    socket[kReset] = true
+  }
+
   if (client[kMaxRequests] && socket[kCounter]++ >= client[kMaxRequests]) {
     socket[kReset] = true
   }
@@ -1378,7 +1398,7 @@ function write (client, request) {
 
   if (upgrade) {
     header += `connection: upgrade\r\nupgrade: ${upgrade}\r\n`
-  } else if (client[kPipelining]) {
+  } else if (client[kPipelining] && !socket[kReset]) {
     header += 'connection: keep-alive\r\n'
   } else {
     header += 'connection: close\r\n'
